/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/


#include <avr/io.h>
#include <avr/signal.h>
#include <inttypes.h>
#include "adc.h"
#include "global.h"
#include "wbo2.h"

// sample and filter analog inputs

uint8_t adc_channel, sample_count;

sample_t sensor_filt[ADC_CHANNELS_AND_BARO];
// sample_t prev_sensor_reading[ADC_CHANNELS];

// initialize the sensor array
   /*!
   * \ingroup g_ad_convert
   *
   * \brief Initialize Analog to Digital Converter.
   *
   * initialize the sensor array
   */
void init_sensors(void) {
	uint8_t i;

	ADCSRA = _BV(ADEN) | ADCSRA_PRESCALER;
  
	for (i=0; i<ADC_CHANNELS; i++){
		sample_t tmp = read_adc(i);
		sensor_filt[i] = tmp + tmp;
	}
}


#ifndef WBO2
   /*!
   * \ingroup g_ad_convert
   *
   * \brief Start Analog to Digital Converter.
   *
   * initilize mux, and set startbit
   */
void startADC(void) {
  // initilize mux, and set startbit

  adc_channel = 0;
  ADMUX = ADMUX_AREF | (adc_channel & 0x07);
#warning "what if the adc hasn't finished?"
  // note, a check is needed such that the adc isn't "overrun"
  ADCSRA = _BV(ADEN) |  _BV(ADSC) | _BV(ADIE) | ADCSRA_PRESCALER;

}
#endif

/* poll adc until conversion is done */
/* used initially for getting the barometric pressure */
   /*!
   * \ingroup g_ad_convert
   *
   * \brief Read Analog to Digital Converter.
   *
   * poll adc until conversion is done
   * used initially for getting the barometric pressure
   * Beware: returns in 10 bit resolution, not 11 as stored in sensor_filt[]
   */
sample_t read_adc(uint8_t channel) {

  /* init adc */
//  ADCSRA = _BV(ADEN) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);  // div-128 prescaler

  // left adjust result
  ADMUX = ADMUX_AREF | (channel & 0x07);
  ADCSRA |= _BV(ADSC);

  loop_until_bit_is_set(ADCSRA, ADIF);
  ADCSRA |= _BV(ADIF); // clear adif-bit

  return ADC;
}




#ifdef WBO2

   /*!
   * \ingroup g_ad_convert
   *
   * \brief Initialize Analog to Digital Converter - WBO2 version.
   *
   * initialize the sensor array
   */
void init_adc(void) {
	adc_channel = 0;
	ADMUX = ADMUX_AREF | (adc_channel & 0x07);
	ADCSRA = _BV(ADEN) | _BV(ADSC) | _BV(ADFR) | _BV(ADIF) | ADCSRA_PRESCALER;
}

   /*!
   * \ingroup g_ad_convert
   *
   * \brief Start Analog to Digital Converter - WBo2 version.
   *
   * initilize mux, and set startbit
   */
void set_adc_channel(uint8_t t) {
	ADMUX = ADMUX_AREF | (t & 0x07);
	ADCSRA |= _BV(ADIF);
}

   /*!
   * \ingroup g_ad_convert
   *
   * \brief Read Analog to Digital Converter - WBO2 version.
   *
   * poll adc until conversion is done
   * used initially for getting the barometric pressure
   */
uint8_t read_adc_channel(uint8_t t) {
	if (ADCSRA & _BV(ADIF)) {
		sample_t tmp = sensor_filt[t];
		// prev_sensor_reading[t] = tmp;
		sensor_filt[t] = (tmp >> 1) + ADC; 
		// shouldn't we clear ADIF by ADCSRA |= _BV(ADIF); ?
		return 1;
	} 

	return 0;
}


/***************************************************************************/
/*** INTERRUPT: ADC conversion complete ***/
/***************************************************************************/

   /*!
   * \ingroup g_ad_convert
   *
   * \brief Analog to Digital Converter Interrupt Handler- WBO2 version.
   *
   * read the converter, set up the next channel
   */
SIGNAL(SIG_ADC) {
	uint8_t sample_idx;
	sample_idx = wbo2_sampling.nernst_sample_cnt;

	// TODO: confirm we do not corrupt memory here (handler2k prevents seeing ff)
	// capture the time after nernst pulsing edge (deliberately always):
	wbo2_sampling.nernst_edgetime[sample_idx] = TCNT2; 

	wbo2_sampling.nernst_sample[sample_idx] = ADC;

	if (sample_idx == WBO2_SAMPLE_COUNT-2) {
		// adc has already started doing the last conversion
		// setup mux for next non-nernst conversion
		set_adc_channel(adc_channel);
		
	} else if (sample_idx >= WBO2_SAMPLE_COUNT-1) {  // done	
		ADCSRA &= ~_BV(ADIE); // disable interrupt
	}

	wbo2_sampling.nernst_sample_cnt = sample_idx + 1;
}


#else	// not WBO2

   /*!
   * \ingroup g_ad_convert
   *
   * \brief Analog to Digital Converter Interrupt Handler.
   *
   * read the converter, set up the next channel
   */
SIGNAL(SIG_ADC) {
	sample_t tmp = sensor_filt[adc_channel];
	// prev_sensor_reading[adc_channel] = tmp;
	sensor_filt[adc_channel] = (tmp >> 1) + ADC; 
	adc_channel++;

  if (adc_channel >= ADC_CHANNELS) { // done
    adc_channel = 0;
    ADCSRA &= ~_BV(ADIE); // disable interrupt
  } else {
    ADMUX = ADMUX_AREF | (adc_channel & 0x07);
    ADCSRA |= _BV(ADSC); // start next conversion ... but why don't we do ADFR ?
  }
}

#endif
