
#ifndef __EVENTQUEUE_H__
#define __EVENTQUEUE_H__

#include "heap.h"
#include <avr/io.h>
#include <stdio.h>
#include <stdarg.h>
/* ---------------------- HARDWARE definitions: 
 */

// EVENTQUEUE_OC2 is not supported, look back older revisions in cvs if you really want
// the 16 below is for "16 bit counter"
#ifdef GENBOARDv3
// OC1C is used from genboard v3.x (economical because of the AVR-pin conflict of OC2 and OC1C)
// TODO: check initialization
#define TCNT16	TCNT1
#define OCR16	OCR1C
#define ICR16   ICR1
#define ETIFR_CLR (_BV(OCF1C))
#define TCCR16B TCCR1B
#define TCCR16B_VAL (_BV(CS11) | _BV(CS10))
#define ETIMSK_OR (_BV(OCIE1C))
#define SIGNAL_OC16 (SIG_OUTPUT_COMPARE1C)
#define SIG_PRIMARY_TRIGGER SIG_INPUT_CAPTURE1
#define SIG_SECONDARY_TRIGGER SIG_INPUT_CAPTURE3
#else 
// OC3C is used on genboard v2.2 since OC1x are for injectors
#define TCNT16	TCNT3
#define OCR16	OCR3C
#define ICR16   ICR3
#define ETIFR_CLR (_BV(OCF3C))
#define TCCR16B TCCR3B
#define TCCR16B_VAL (_BV(CS31) | _BV(CS30))
#define ETIMSK_OR (_BV(OCIE3C))
#define SIGNAL_OC16 (SIG_OUTPUT_COMPARE3C)
#define SIG_PRIMARY_TRIGGER SIG_INPUT_CAPTURE3
#define SIG_ALIEN_MEASURE SIG_INTERRUPT6
#endif

/* we take statistics for this many pulses + 1 */
#define ROTBY	2
#define AGGRNUM	(1)


/* We always schedule a handler with a fix frequency.
H2KDIVIDE obsoleted:
we schedule handler2k() every time when TCNT2 overflows. But we do not use
TOV2, just OCR2 interrupt.
obsolete:
 * This way we get rid of the OCR0 handler, so TIMER0 can be used for anything.
 * 250: would be 1000Hz but that's playing with fire (if latency is too high, more
 * than 256 counts happen. We could use the fact that a minimal amount of time
 * surely passed. Some heuristics could make this possible. I'm afraid of this...
 * 200: 800 usec = 1250Hz
 * 125: 500 usec = 2000Hz (the name handler2k comes from this) 
 * seem good values, but you need to set the corresponding 
 * divisors accordingly so the lower freqs are not screwed.
 */
// #define H2KDIVIDE 200

/* In case we have nothing to do for quite a while, the next OCR value will be set (now - OCR_SAFETY):
 * in practice delay will be (256 - OCR_SAFETY) (TCNT2 is 8 bit) 
 * If OCR_SAFETY value is small, the exec-delay of OCR interrupt might result in
 * TCNT2 reach same or possibly higher value than read before
 * in readTMR2() => higher timer bits not updated
 * correctly. You can set this lower if you know all the interrupt and cli()-sei() times, but better be safe.
 * Remember that with a small value you gain little and risk a lot.
 * OBSOLETE
 */
// #define OCR_SAFETY 10

// for persum (sum of periods) ROTBY is not used 
#define RPM(A) (250000 * 60 * AGGRNUM / INJ_PER_REV /A)

/* refuse periods smaller than 2000 usec (500 * 4usec)
 * this means 500 Hz, or 30000 events/min, 15000RPM 
 * NOT USED IN IC3
 */ 
#define MINPER 500

/* before the injector opens, it fires appr. 1/3 of nominal rate
 * This is the correction in 4usec cycles :
 * eg. 10 means 40usec, appropriate where injector opens in 120 usec 
 */
#define INJ_OPEN_CORR	10

struct tachin_t {
	uint32_t	pwsum;
	uint32_t	pwsqrsum;
	uint32_t	persum;
	uint32_t	persqrsum;
	uint16_t	rise_0;
	uint16_t	rise_prev;
	uint8_t		edge_cnt;
	uint8_t		edge_bad;
};

/* updated when calc ready, used by display */
struct pwdisplay_t {
	uint16_t	pwsum;
	uint16_t	pwsqrsum;
	uint16_t	persqrsum;
	uint32_t	persum;
	uint16_t	rpm;
	uint8_t		bad;
	uint8_t		dirty;
#ifdef PHASESIGNAL
	uint8_t		toothcount;
#endif
};

extern struct tachin_t tachin;
extern struct pwdisplay_t pwdisplay;


struct etime_t {
uint16_t now;
uint16_t now_upper;
uint8_t	h2klast;
// uint16_t now_follow;	// OCR2 interrupt always comes _after_ the time setalarm has scheduled
uint8_t h2klastrunat;		// the now upper byte when h2k was last executed
uint32_t ptrig_capturedtime_prev; // TODO: maybe more historic values?
uint32_t strig_capturedtime_prev; // secondary trigger
uint8_t altcount; // Alternate count selector
uint8_t trigcount; // trigger counter for injection scheduling
uint8_t egocount;
uint8_t asecount;
uint8_t ase_cyl_count;  // this will leave when we keep track of the engine cycle

uint16_t strig_capturedtime_dwell;

	uint8_t tooth_cnt;	// tooth wheel in sync when etime.tooth_cnt < 128. Initial sync sequence is downcounting from 255
	uint32_t prev_missing_tooth;
	uint16_t average_tooth_period;

	uint32_t primary_trigger;

  uint8_t ego_warmup_cnt;
  uint8_t wbo2_sampling;
	uint8_t cam_sync_cnt;
	uint8_t cam_trigger_expire;
};

extern struct etime_t etime;
// #define ho etime.heap_overflow;
// #define hn etime.heap_normal;
// #define now (etime.now);

/* include "eventqueue.c", these are not needed. Looks like a hack to include the 
C file, but still better then put them together into one file, or have lotsof
function-calls. I tried several things here, but only static inline worked (maybe my fault)
http://gcc.gnu.org/onlinedocs/gcc-3.3.1/gcc/Inline.html#Inline
*/
void eventqueue_init(void);
void tachin_reset(void);
void pwdisplay_from_tachin(uint16_t);
void event_insert(key_t, val_t);
void event_insert_destroy(key_t k, val_t v);
void event_insert_rel(key_t dif, val_t v);
void event_session_start(void);
void event_insert_takecare(key_t k, val_t v);
void event_session_end(void);
void alarm2core(void);
void counter_sync(void);
/*
void dispatcher(val_t v);
uint8_t pw2comm(uint8_t);
uint8_t pwdata(uint8_t);
*/

struct heap_t h1;
#define heap_normal (&h1)
// #define event_action_required(x) ((int16_t)(x-etime.now) <= 0)
#define EVENT_MOST_FUTURE (32700)
// TODO: now_upper:
static inline key_t readTMR(void) {
	key_t newtime, oldtime = etime.now;
	etime.now = newtime = TCNT16; 
	if(newtime < oldtime) // note: faster and better than earlier16()
		etime.now_upper++;
	
	return newtime;
} // readTMR()
	
#endif
