   /*!
   * \addtogroup g_keyboard Keyboard Control Routines
   */
/*@{*/
#ifdef PS2KEYBOARD

/* written by Marcell Gal, cell at x-dsl.hu
 * original source is from AVR313 appnote
*/

#include "kb.h"
#include "kbpindefs.h"
#include "global.h"
#include "menu.h"
#include "eventqueue.h"
#include "benchmark.h"
#include "comm.h"
#include "lcd.h"
#include "scancodes.h"

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/signal.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/delay.h>

#define BUFF_SIZE 32

uint16_t parity_ok, parity_bad;

unsigned char bitcount;                // 0 = neg.  1 = pos.

// We put raw data into the buffer (unlike AVR313 appnote)
// and running decode from userspace (non-irq)
unsigned char kb_buffer[BUFF_SIZE];
unsigned char *inpt, *outpt;
volatile unsigned char buffcnt;		// the number of chars in the buff at the moment

void init_kb(void)
{
    inpt =  kb_buffer;                        // Initialize buffer
    outpt = kb_buffer;
    buffcnt = 0;

    bitcount = 11;
    kb_lock = 0;

// Both data and clock are used as normal input port, nothing to init there

#ifdef GENBOARDv3

#if CLOCK == 6
    EICRB |= _BV(ISC61);	// page 86.
	EICRB &= ~_BV(ISC60);	// Set interrupt on falling edge
#else
#warning "YOU WILL NOT GET PS2 CLOCK INTERRUPT"
#endif
    EIFR = _BV(INTF6);		// clear interrupt flag
    EIMSK |= _BV(INT6);		// page 88. enable interrupt

#else	// GENBOARD v2.2

#if CLOCK == 2
    EICRA |= _BV(ISC21);	// page 86.
	EICRA &= ~_BV(ISC20);	// Set interrupt on falling edge
    //    EICRA |= _BV(ISC20);		// page 87. Set interrupt on rising edge
#else
#warning "YOU WILL NOT GET PS2 CLOCK INTERRUPT"
#endif
    EIFR = _BV(INTF2);		// clear interrupt flag
    EIMSK |= _BV(INT2);		// page 88. enable interrupt
#endif	// GENBOARD v2.2
}

// only called from interrupt
__inline__ void put_kbbuff(unsigned char c)
{
    if (buffcnt<BUFF_SIZE)                        // If buffer not full
    {
        *inpt = c;                                // Put character into buffer
        inpt++;                                    // Increment pointer

        buffcnt++;				// no need to cli(), since we are in irq

        if (inpt >= kb_buffer + BUFF_SIZE)        // Pointer wrapping
            inpt = kb_buffer;
    }
}

__inline__ void grab_kb_clock(void) {
	PORT_KB &= ~_BV(CLOCK); // set signal low
	DDR_KB |= _BV(CLOCK);  // make it an output
	kb_lock = 2;  // stay in this state for at least 100 us
}

__inline__ void release_kb_clock(void) {
	DDR_KB &= ~_BV(CLOCK);  // make signal an input
	PORT_KB |= _BV(CLOCK);  // enable pull-up
}


// clock-edge interrupt
// Check the time and reset bitcount of too much time passed
// This may not recover safely from lost sync (better than avr313 appnote,
// which lacks this alltogether). Specifications should be checked
SIGNAL(PS2_INTERRUPT){
    static unsigned char data;                // Holds the received scan code
	static unsigned int tm_prev;
	
#ifdef BENCHMARK
	benchstats.kbirqcnt++;
#endif
	// check the time (now) and start again if too much elapsed since last time
	readTMR();
	if (etime.now_upper - tm_prev > 1){		// more than 1/4 sec passed (sloppy)
/* MIK:
   The kb-clock recovery isn't strictly necessary right here.
   If the keyboard hasn't sent anything for 1/k second, then this falling clock
   edge is propably the start of a new transmission.

		// try recovering from the transmission error
		grab_kb_clock(); // tell kb to stop sending more bits
				// kb will resend the broken scancode when clock is released
//		put_kbbuff(0xff); // magic "scanbyte" == clear flags in decode()
		bitcount = 11+1; //+1 since we subtract 1 below
*/
		bitcount=11;
#ifdef BENCHMARK
		benchstats.kbresync++;
#endif
	}
	
	tm_prev = etime.now_upper;

	if(bitcount < 11 && bitcount > 2)    // Bit 3 to 10 is data. Parity bit,
        {                                    // start and stop bits are ignored.
            data = (data >> 1);			// LSB comes first
            if(PIN_KB & _BV(DATAPIN))
                data = data | 0x80;            // Store a '1'

	} else if (bitcount == 11) { // the start bit
		if (PIN_KB & _BV(DATAPIN)) {  // check if start-bit ain't low
			parity_bad++;

                        // try recovering from the transmission error
			grab_kb_clock(); // tell kb to stop sending more bits
					// kb will resend the broken scancode when clock is released
			put_kbbuff(0xff); // magic "scanbyte" == clear flags in decode()
			bitcount = 11+1; //+1 since we subtract 1 below
		}

	} else if (bitcount == 2) { // the partity bit
		uint8_t i, pdat=0;

		for (i=0; i<8;i++) // count numbers of set bits in the received byte
			if (data&(1<<i)) pdat++;

		if ((PIN_KB & _BV(DATAPIN)) && (pdat&0x01)) { // odd-parity
			parity_bad++;

                        // try recovering from the transmission error
			grab_kb_clock(); // tell kb to stop sending more bits
					// kb will resend the broken scancode when clock is released
			put_kbbuff(0xff); // magic "scanbyte" == clear flags in decode()
			bitcount = 11+1; //+1 since we subtract 1 below

		} else {
			parity_ok++;
		}
	}

        if(--bitcount == 0)                    // All bits received
        {
            put_kbbuff(data);			// we put raw value into buffer
            bitcount = 11;
        }
   
}

/* lower byte is 0 for special keys.
TODO: map special keys (F keys!!) to the lower byte so that usable from menu.c
*/
unsigned int decode(unsigned char sc)
{
	static unsigned char kbflags = 0; //, mode = 0;

	if (debug_ps2 & 0x01) {
		static uint8_t tmp;
		tmp=sc;
		pushfuncpar(&srl_generic, 1, &tmp);
	}
	if (sc == 0xff) { // on second thought, is this necessary?
		kbflags &= ~_BV(is_e0); // clear multichar related flags;
		return (sc<<8);
	}
	
	if ((kbflags & _BV(is_up)) == 0)                // Last data received was NOT the release-key identifier
	{
		switch (sc)
		{
		case 0xF0 :        // The up-key identifier
			kbflags |= _BV(is_up);
			break;

		case 0xE0 :
			kbflags |= _BV(is_e0);
			break;

		case 0x12 :        // Left SHIFT
			kbflags |= _BV(lshift);
			break;

		case 0x59 :        // Right SHIFT
			kbflags |= _BV(rshift);
			break;

		default:
			if(kbflags & _BV(is_e0)){
				kbflags &= ~_BV(is_e0);
				switch(sc){
				case 0x70: return KEY_INS;
				case 0x6c: return KEY_HOME;
				case 0x7d: return KEY_PGUP;
				case 0x71: return KEY_DEL;
				case 0x69: return KEY_END;
				case 0x7a: return KEY_PGDN;
				case 0x75: return KEY_UPARROW;
				case 0x6b: return KEY_LEFTARROW;
				case 0x72: return KEY_DNARROW;
				case 0x74: return KEY_RIGHTARROW;
				case 0x4a: return KEY_GRSLASH;
				case 0x5a: return 13;	// grey enter: same as the main enter
					// 0x14 and 0xe014 (left and right control) are not handled
				default: return 0;
				}

			} else {
				
				if(sc&0x80) {		// too big, our tables are 0..7F
					if (sc==0x83)
						return KEY_F7; // last chance!
					else
						return sc<<8;
				}

				if(kbflags & (_BV(rshift) |_BV(lshift))) // If left or right shift pressed,
				{                            // do a table look-up                
					return(PRG_RDB(shifted + sc));             
				} else {                    // no shift pressed
					return(PRG_RDB(unshifted + sc));
				}
			}	// not after e0
			break;
		}
	} else {
		kbflags &= ~_BV(is_up);                            // Two 0xF0 in a row not allowed
		kbflags &= ~_BV(is_e0);  // release of extended key sends 0xE0 0xF0 0xNN
		switch (sc)
		{
		case 0x12 :                        // Left SHIFT
			kbflags &= ~_BV(lshift);
			break;
			
		case 0x59 :                        // Right SHIFT
			kbflags &= ~_BV(rshift);
			break;
		} 
	}
	return sc<<8;	// some modifier key
} 


// call from userspace mainloop (return immediately, no busywait)
void ps2proc_char(void)
{
	unsigned char t;
	while(buffcnt) {	// there is no upper limit for number of chars processed
	t = *outpt++;
	if (outpt >= kb_buffer + BUFF_SIZE)            // Pointer wrapping
		outpt = kb_buffer;
    cli();
	buffcnt--;		// this is also read and written by interrupt
	sei();
	t= decode(t) & 0xFF;
	if(t) {
		if (debug_ps2 & 0x02) {
			static uint8_t tmp;
			tmp=t;
			pushfuncpar(&srl_generic, 1, &tmp);
		}
		menu_char(&menu_ps2, t);
	}
	}
}

// this busywaits (no good for production)
int ps2getchar(void)
{
    int byte;
    while(buffcnt == 0);                        // Wait for data

    byte = *outpt;                                // Get byte
    outpt++;                                    // Increment pointer

    if (outpt >= kb_buffer + BUFF_SIZE)            // Pointer wrapping
        outpt = kb_buffer;
    
	// looks like a locking error in the appnote
    buffcnt--;                                    // Decrement buffer count

    return byte;
}

#ifdef BENCHMARK
uint8_t kbuf_comm(uint8_t i)
{
	switch(i){
		case 0: return hex_hi(benchstats.kbcharlast>>8);
		case 1: return hex_lo(benchstats.kbcharlast>>8);
		case 2: return hex_hi(benchstats.kbcharlast);
		case 3: return hex_lo(benchstats.kbcharlast);
		case 4: return 'k';
		default: return 'p';
	}
}		

#endif // BENCHMARK

uint8_t ps2_stat_comm(uint8_t i)
{
	switch(i){
		case 0: return '\r';
		case 1: return '\n';
		case 2: return ' ';
		case 3: return 'K';
	        case 4: return 'B';
	        case 5: return ':';
	        case 6: return ' ';
		case 7: return hex_hi(parity_ok>>8);
		case 8: return hex_lo(parity_ok>>8);
		case 9: return hex_hi(parity_ok&0x00FF);
		case 10: return hex_lo(parity_ok&0x00FF);
	        case 11: return ' ';
		case 12: return hex_hi(parity_bad>>8);
		case 13: return hex_lo(parity_bad>>8);
		case 14: return hex_hi(parity_bad&0x00FF);
		case 15: return hex_lo(parity_bad&0x00FF);
		default: return 'e';
	}
}


#endif // PS2KEYBOARD

/*@}*/

