
#ifdef GENBOARDv3 // should be KNOCK_8101, but lotsof references without ifdef

#include <inttypes.h>
#include <avr/io.h>
#include <avr/signal.h>
#include <avr/sfr_defs.h>
#include <avr/interrupt.h>

#include "arithm.c"

#include "global.h"
#include "knock.h"
#include "spi.h"
#include "ve.h"
#include "eventqueue.h"
#include "benchmark.h"
#include "comm.h"
#include "dispatcher.h"
#include "lbus.c"

struct knock_t knock;
struct knock_cyl_t knock_cyl[KNOCK_CHAN];
struct knock_stats_t knock_stats;

void reset_knock_adjust(void) {
	uint8_t i;
	for (i=0; i<KNOCK_CHAN; i++) {
		knock_cyl[i].knock_adjust = config.knock_default_retard;
	}
}

void init_knock(void) {
	uint8_t i;

	for (i=0; i<KNOCK_CHAN; i++) {
		knock_cyl[i].knock_adjust = config.knock_default_retard;
		knock_cyl[i].state = K_NO_KNOCK;
		knock_cyl[i].cnt = 0;
		knock_cyl[i].kthres_high_cnt = 0;
		knock_cyl[i].kthres_low_cnt = 0;
	}

	knock.igncount = 0;

	knock_stats.count = 0;
	knock_stats.pos_intensity_cnt = 0;
	knock_stats.neg_intensity_cnt = 0;
	knock_set_state(KNOCK_IDLE);
	knock.status &= ~_BV(knock_window_valid);
}


/* retarding timing a bit [currently 1/4 crankdegrees] */
static inline uint8_t knock_retard(uint8_t old, uint8_t how_much)
{

	uint8_t ret, max;
	ret = old + how_much;
	max = config.knock_max_retard;

	if(ret <= max)
		return ret;

	return max;
}

/* advance timing a bit [currently 1/4 crankdegrees] */
static inline uint8_t knock_advance(uint8_t old, uint8_t how_much)
{
	uint8_t ret;
	ret = old - how_much;
	if(ret <= old) return ret;

	return 0;	// instead of overflow
}

/* 
  decide which channel to sample next

  with a V-engine, there should be a table that specifies which knock
  channel to use as function of engine.igncount
*/
uint8_t next_knock_channel(void) {
	if (config.knock_conf & _BV(knock_chan)) 
		return KNOCK_CHANNEL_2;
	else 
		return KNOCK_CHANNEL_1;
}

/*
  scale the length of the knock sampling window to
  obtain the length of the noise sampling window

  4 cyl: 720 degrees / 4 = 180 degrees
  6 cyl: 720 degrees / 6 = 120 degrees
  8 cyl: 720 degrees / 8 = 90 degrees

  knock window + noise window + margin < total degrees per ignition event
  ...where margin is due to the asynchronous reading of spi data

  4 cyl is easy, the noise sample window can be the same length as
    the knock sampling window.
  8 cyl...hmmm, with a max knock sampling window of 64 degrees, there
    is not much time left. 
	Even if swapping between the banks (V8 needs 2 sensors), sometimes
	there is only 90degrees between events on the same microphone.
	Mustangs 289ci V-8 engine, for example, has a firing order of 
	1-5-4-2-6-3-7-8.
	other:  
	1 5 4 8 6 3 7 2
	1-6-2-5-8-3-7-4
	1 8 4 3 6 5 7 2

	So when 2 neighbor cyls fire on the same bank, 
	there is little time to listen for noise.
  10+ cyl, gotta be tough

  for 4 cyl: noise window == knock window / 1
  for 8 cyl: noise window == knock window / 4
*/

// downscale the length of the knock sampling window and return 
// the length of the noise sampling window
static inline uint16_t noise_sampling_window(void)
{
	uint8_t ncyl, scale;
// TODO: unconvolute this, what if cyl>8?
	ncyl = NCYL_1;
	if (ncyl < 3)
		ncyl = 3;

	scale = (7-ncyl) * 3 + 4;
	return mult16_8(knock.window_length, scale) >> 4;
}

/* compensate for reduced sampling window,
* eg. the sampled background noise is up-scaled.
* examples:
* ncyl sampling_window unitscale multiple
 3	16	16	256
 4	13	28	364
 5	10	40	400
 6	7	52	364
 7	4	64	256
 TODO: The above can be acceptable, if thresholds are configurable,
 but do we just want a flash-lookup instead ?
*/
static inline uint16_t noise_val_unitscale(uint16_t v)
{
	uint8_t ncyl, scale;

	ncyl = NCYL_1;
	if (ncyl < 3)
		ncyl = 3;

	scale = (ncyl-3) * 12 + 16;
	return mult16_8(v, scale) >> 4;
}

/* logs knock samples
	BEWARE: only call this if there is some place in cmd buffer
*/
void knock_detailed_log(struct knock_cyl_t *kc)
{
	uint8_t *comm_TX_plainbytesP;
#ifdef KNOCK_8101
	uint8_t indx;
	uint8_t *p;
	uint16_t t16;

	comm_TX_plainbytesP = comm_TX_plainbytes;
	// let the sample start with "\r\nk="
	* comm_TX_plainbytesP++ = '\r';
	* comm_TX_plainbytesP++ = '\n';
	* comm_TX_plainbytesP++ = 'k';
	* comm_TX_plainbytesP++ = '=';

	* comm_TX_plainbytesP++ = knock.igncount; 
	* comm_TX_plainbytesP++ = engine.ignadv;

	// rpm and kpa are nice in the log:
	t16 = engine.rpm_vhr;
	*comm_TX_plainbytesP++ = t16 >> 8;
	*comm_TX_plainbytesP++ = (uint8_t)(t16);

	t16 = engine.kpa250;
	*comm_TX_plainbytesP++ = t16 >> 8;
	*comm_TX_plainbytesP++ = (uint8_t)(t16);
	
	/* anything else to log ? */

	// push a sample-set:
	// log whole struct knock_cyl_t
	// NOTE, that we could pushfuncpar() the kc pointer (losing atomicity)
	p = ((uint8_t *) kc ) + sizeof (struct knock_cyl_t );
	comm_TX_plainbytesP += sizeof (struct knock_cyl_t );
	indx = sizeof (struct knock_cyl_t );
	while(indx){
		* --comm_TX_plainbytesP = * --p;
		--indx;
	}

	// do we want several _consequtive_ sample-sets ? 
	// This way because of slow RS232 maybe there are no 2 neighboring sets..

	// calculate log size: 4(header) +1(ki) +1(ign_adv) +2(rpm_vhr) +2(kpa250)
#define KNOCK_DETAILED_OVERHEAD (4+1+1+2+2)
#define KNOCK_LOG_TOTAL_LEN (sizeof (struct knock_cyl_t ) + KNOCK_DETAILED_OVERHEAD)
	pushfuncpar(NULL, KNOCK_LOG_TOTAL_LEN, comm_TX_plainbytes);
#endif // KNOCK_8101

}	// knock_detailed_log

/* evaluate the result
 advance = n - adjust
 no-knock: adjust = 0 always, simple
 knock-detection:
 
 adjust = config.knock_default_retard initially for knock 
 (min.range: 0..16 degrees, 1 degree resolution; currently better).
 default retard is reduced slowly to 0 if knock sensor is happy.
 increased relatively fast, possibly railing at 
 config.knock_max_retard if knock happens.
 
 config.knock_max_retard could be replaced by
 config.knock_default_retard + added_retard_limit 
 reasonable range for added_retard_limit is 0..16 (maybe 32? nay)
 crankdegrees, 1 degree step would be OK; currently much better
*/

//currently the same knock_adjust is used for entire rpm x map, is
//this good?
void knock_act(struct knock_cyl_t *kc) {
	uint8_t ka;
	// uint8_t kth, ktl;
	uint8_t knock_state;
	uint16_t nsignal;
	uint16_t kv, nv;

	if(((debug.enable_flags & _BV(2))) // don't enable mde.. bit0 at the same time !
	&& (*comm_TX_plainbytes == 0)	// the signalling byte is already zeroed by comm.c (gcc-at-powerup)
	&& ! comm_buffer_full()){
		knock_detailed_log(kc);
	}

	kv = kc->knock_val;
	nv = kc->noise_val;
	// collect some stats of the data (debug only)
	if (kv > nv)
		knock_stats.pos_intensity_cnt++;
	else
		knock_stats.neg_intensity_cnt++;

	// scale noise_val to determine if the knock_val contains knocking:
	// knock_val > noise_val * scale + threshold

	// the >>4 is arbitrary, any better suggestions?
	nsignal = mult16_8(nv, config.knock_noise_scale) >> 4;

	// not sure if there is a need for the <<2,
	// since it is very unlikely that noise_val is zero
	nsignal += config.knock_threshold << 2;  // knock.data is 10 bit

 	// determine if there are knocking or not
	if (kv > nsignal) {
		knock_state = K_KNOCK;
		// remember how many times knock occured
		if (kc->kthres_high_cnt < 0xff) kc->kthres_high_cnt++;
	} else {
		knock_state = K_NO_KNOCK;
		if (kc->kthres_low_cnt < 0xff) kc->kthres_low_cnt++;
	}

	if (kc->cnt) {
		return;  // do nothing until kc->cnt has expired
	}


	// this part isn't used at the moment
	if (knock_state == kc->state) {
		// time spent in this state is 
		// kc->kthres_high_cnt + kc->kthres_low_cnt
	} else {
		kc->kthres_high_cnt = 0;
		kc->kthres_low_cnt = 0;
	}

	kc->state = knock_state;
	ka = kc->knock_adjust;

	// finally knock_adjust is modified
	switch(knock_state) {
	case K_KNOCK:
		ka = knock_retard(ka, config.knock_retard_step);
		kc->cnt = config.knock_retard_delay;
		break;

	default:// K_NO_KNOCK
		// normal behavior, no knocking
		ka = knock_advance(ka, config.knock_advance_step);
		kc->cnt = config.knock_advance_delay;
		break;
	}

	kc->knock_adjust = ka;
}


/*
  This function brings knock_adjust towards config.knock_default_retard

  When there is no knock-evaluation, the knock_adjust should instantly
  increase to default_retard, or slowly decrease to default_retard

 */
void knock_act_default(struct knock_cyl_t *kc) {
	uint8_t ka, kdr;

	ka = kc->knock_adjust;
	kdr = config.knock_default_retard;

	kc->state = K_NO_KNOCK;

	if (ka == kdr)  // no probs, nothing to do
		return;

	// the easy case: retard the ignition _fast_
	if (ka < kdr) {
		kc->knock_adjust = kdr;
		return;
	}

	// ka > kdr, advance slowly
	if (kc->cnt)  // do nothing until timer has expired
		return;

	// slowly advance ignition until knock_default_retard is reached
	ka = knock_advance(ka, config.knock_advance_step);
	kc->cnt = config.knock_advance_delay;

	if (ka < kdr)  // dont advance too much!
		ka = kdr;
		
	kc->knock_adjust = ka;

}



// time based knock delay
// decrease the counter of each channel
void decrease_knock_cnt(void) {
	uint8_t i;
	for (i=0; i<KNOCK_CHAN; i++) {
		if (knock_cyl[i].cnt)
			knock_cyl[i].cnt--;
	}
}

/*
  Take care of communication with dispatcher and spi-handler, and make
  sure the sampled knock data are stored where it belongs
*/
void handle_knock(void) {
	uint8_t ki, state;
	uint16_t length;
	ki = knock.igncount;
	state = knock_state();

	switch(state) {
	case KNOCK_WINDOW_START:
		// disp_knock_stop() changes the state to KNOCK_WINDOW_DONE
		return;

	case KNOCK_WINDOW_DONE:
		// read_knock() is kicked off to start the pickup of the data. 
		read_knock();   // request retrieval of data
		knock_stats.count++;  // for debugging purposes
		cli();
		knock_nextstate();  // goto NOISE_WINDOW_PICKUP
		sei();
		return;

	case KNOCK_WINDOW_PICKUP:
		// check if data has arrived
		if (knock.data == 0xffff) return;
		
		knock_cyl[ki].knock_val = knock.data;

		cli();
		knock_nextstate();  // goto NOISE_WINDOW_START
		sei();

		// then schedule KNOCK_STOP event, such that the
		// background noise can be obtained

		// scale knock sampling window to get noise sampling window
		length = noise_sampling_window();

		// manually start sampling the knock sensor and insert stop-event
		cli();
		readTMR();
		PORTC |= _BV(KNOCK_INTEGRATE);
		event_insert(etime.now + length, KNOCK_STOP);
		if (!(length>>8)) // handler2k will fire in max 256*4us and run alarm2core()
			alarm2core();
		sei();

		//TODO: calculate next knock_channel (will
		//      automatically be configured during read_knock()
		return;

	case NOISE_WINDOW_START:
		// disp_knock_stop() changes the state to NOISE_WINDOW_DONE
		return;

	case NOISE_WINDOW_DONE:
		read_knock();   // request retrieval of data
		cli();
		knock_nextstate();  // goto NOISE_WINDOW_PICKUP
		sei();
		return;

	case NOISE_WINDOW_PICKUP:
		// check if data has arrived
		if (knock.data == 0xffff) return;

		// not knock-value, then it must be the background noise
		knock_cyl[ki].noise_val = noise_val_unitscale(knock.data);

		// if parameters are going to be set, then now is a good time.
		// remember that the knock IC ignores spi activity while
		// KNOCK_INTEGRATE is high, thus in case of a race, the
		// parameters might be ignored, what can we do about that?
		set_knock_channel(next_knock_channel());
		
		cli();
		knock_set_state(KNOCK_IDLE);
		sei();
		break;  // NOTE: we dont return, but continue 

	default:  // KNOCK_IDLE
		return;
	}

	// check if there was response from the knock chip
	if (knock.status & _BV(KNOCK_NOT_RESPONDING)) {
		knock_act_default(&knock_cyl[ki]);
		return;
	}

	// check if rpm is within range for trustful knock detection
	// TODO: maybe add some hysteresis
	if ((engine.rpm < config.knock_minrpm) || (engine.rpm > config.knock_maxrpm)) {
		knock_act_default(&knock_cyl[ki]);
		return;
	}

	// act on the data
	knock_act(&knock_cyl[ki]);
}


// helper function for calc_knock_window
uint16_t k_window(uint16_t t) {
	uint32_t t32;
	uint16_t per;

	per = (uint16_t)engine.rpm_period;
	t32 = mult16_16(per, degree2time(t));
	if(lsr24(t32) < 16)
		return div4096(t32);
	
        // rpm too low, we dont want knock, signal with ffff
	return 0xffff;
}


// calculate when the knock window starts/ends
// conditions that prevents enabling of knock can be inserted in this function
void calc_knock_window(void) {
	uint16_t start, length, per;

	if (!(config.knock_conf & _BV(knock_enable))) {
		cli();
		knock.status &= ~_BV(knock_window_valid);
		sei();
		knock.data = 0xffff;
		reset_knock_adjust();  // knock control disabled, default retard
		return;
	}

	// calculate the window timing etc.
	per = engine.rpm_period >> 16;
	if (per == 0) {
		uint8_t ksw;
		ksw = config.knock_sampling_window;

		// turn 4 degree resolution into 0.25 degree
		// the KNOCK_START event will be inserted from IGNDEACT (at ignadv/4 degrees BTDC)
		start = k_window(engine.ignadv + ((ksw & 0x0f) << 4));

		// turn 4 degree resolution into 0.25 degree (happens automatically)
		length = k_window(ksw & 0xf0);

	} else {
		start = 0xffff;
		length = 0xffff;
	}

	cli();
	knock.window_start = start;
	knock.window_length = length;
	sei();


	// set a flag that controls dispatcher-insertion of the knock event
	if ( (start != 0xffff) && (length != 0xffff) ) {
		cli();
		knock.status |= _BV(knock_window_valid);
		sei();
	} else {
		cli();
		knock.status &= ~_BV(knock_window_valid);
		sei();

	        // for some reason the knock window was not
	        // calculated, reset knock_adjust
		knock_act_default(&knock_cyl[engine.igncount]);
	}
}


#endif // GENBOARDv3
