
/*
--- HD44780 compatible LCD display support ---

On GenBoard the LCD is connected with the databus[7:4] on the high bits of PORTA (these can be 'tapped'
at the four smd resistors [R4:R1] going to the four LEDs). The three control signals resides on
PORTD and PORTG are available on the 5x2 header connector.
AfreshBoard pinout is somewhat different, see below.

TIP: If you have a display that requires a negative voltage supply for the LCD driver (or just
have a weak contrast), -9V can be found at capacitor C4 next to the max232.
*/



#ifndef __LCD_H__
#define __LCD_H__

#include <avr/io.h>
#include <stdio.h>
#include <stdarg.h>

/* ---------------------- HARDWARE definitions: 
 * data lines are higher 4 bits(0xF0)
 * This is currently wired in lcd_cmd() and lcd_data()
 */

#ifdef JAPANESE_LCD
// our Japanese friends use a different charset
#define DEGREESYMBOL	0xDF
#else
#define DEGREESYMBOL	0xB0
#endif 

/* any symbol will do: */
#define OUT_OF_RANGE_CHAR 0xAD

#ifdef LCD_2x16

#define LCD_LINES	2
#define LCD_COLS	16

#elif LCD_2x20

#define LCD_LINES	2
#define LCD_COLS	20

#elif LCD_4x20

#define LCD_LINES	4
#define LCD_COLS	20

#elif LCD_4x16

#define LCD_LINES	4
#define LCD_COLS	16

#elif LCD_1x16

#define LCD_LINES	1
#define LCD_COLS	16

#else

#warning "No LCD size defined, check my_make"

#endif

#define LCD_2ND_LINE_ADDR 0x40
#define LCD_CHARS	(LCD_LINES*LCD_COLS)

//------------------
// the master LCD_MAPPING selection switch:
#ifndef LCD_MAPPING

#ifdef GENBOARD
#define LCD_MAPPING 2
#endif

#ifdef GENBOARDv3
#define LCD_MAPPING 3
#endif

#ifdef AFRESHBOARD
#define LCD_MAPPING 3
#endif

#endif	// LCD_MAPPING

//--- it is possible to define all in my_make if you like
#ifndef LCD_DATA

#if LCD_MAPPING==3
#define LCD_DATA        PORTC
#define LCD_DATAPIN     PINC
#define LCD_DATADDR     DDRC
#define LCD_DATADDR30	(0x0F)
#define LCD_CTRL_EN     PORTC
#define LCD_CTRLDDR_EN  DDRC

#define ENABLE  _BV(1)

/* direction: 0: write, 1:read (purple) */
#define DDIR   _BV(0)

/* RS=0: instruction, RS=1: data
	RS is fed from s259 out0 */
#define LCD_RS_259_PORT	(CS_S259_PORT)
#define LCD_RS_259	(CS_S259)
#define LCD_RS	(0)

#define LCD_CTRL_RS_RW     PORTC
#define LCD_CTRLDDR_RS_RW  DDRC

#else // LCD_MAPPING==2
// The lcd is connected to the upper 4 bits of porta (the same signals as the leds)
#define LCD_DATA        PORTA
#define LCD_DATAPIN     PINA
#define LCD_DATADDR     DDRA
#define LCD_DATADDR30	(0x08)
#define LCD_CTRL_EN     PORTD
#define LCD_CTRLDDR_EN  DDRD 

#define ENABLE  _BV(4)
/* RS=0: instruction, RS=1: data (grey) */
#define RS      _BV(1)
/* direction: 0: write, 1:read (purple) */
#define DDIR   _BV(0)

#define LCD_CTRL_RS_RW     PORTG
#define LCD_CTRLDDR_RS_RW  DDRG 

#endif	// LCD_MAPPING

#endif	// LCD_DATA

/* ------------------- END of HARDWARE definitions */

typedef unsigned char byte;
typedef uint8_t lcdpos_t;

uint8_t commbuffer[10];

struct lcd_t {
uint8_t init;	// zero when normal operation. Init-sequence otherwise
char 	cache[LCD_CHARS];/* we keep an in memory copy of the LCD data, then update one at a time */
uint8_t	dirty[LCD_CHARS/8];
/* to avoid lotsof lcd_goto commands we try to keep track of where the cursor is.
 * NOTE: This is index in lcd_cache, not address in CGRAM  */
lcdpos_t went_to;

/* what should be checked next: */
lcdpos_t pos_to_peek;

/* position in cache that will be written next time by eg. printhex */
lcdpos_t next_pos;

/* sometimes the full cache should be set dirty and maybe rarely an init would be good? */
 uint16_t lowerfreq; 
};

extern struct lcd_t lcd;

void lcd_init_seq(void);
__inline__ void lcd_dirty(void);
void lcd_delay(uint16_t p);
void toggle_E(void);
void lcd_cls(void);
void lcd_home(void);
void lcd_control(byte disonoff, byte curonoff, byte curblink);
void lcd_goto(byte mesto);
void lcd_putch(byte data);
void lcd_putstr(byte *data);
void lcd_clr_to_end(void);
void lcd_clr_to_idx(uint8_t idx);
void lcd_clr_rest_of_line(void);
void printbin(byte x);
void printhex(byte i);
void print2(uint8_t x);
void print3(uint16_t x);
void print5(unsigned int x);
void print10(unsigned long x);
void lcd_init(void);
void lcd_initfrom(uint8_t startcountfrom);
void lcd_testLOW(void);
void lcd_testHIGH(void);
// void lcd_testrun(void);
void lcd_start_address_test(void);
void lcd_charseq_test(uint8_t);
uint8_t lcd_busy(void);
void lcd_refresh(void);
void int2BCD5(uint16_t bin, lcdpos_t index);

#define BCD_DIGIT4	2
#define BCD_DIGIT3	1
#define BCD_DIGIT2	0

// should be close to 0xFF (test will be 2.5 sec * distance from 0xFF long)
#define LCD_ADDRESS_TEST 0xFC
#define LCD_INIT_STATE (LCD_ADDRESS_TEST - 10)

#define FLASH __attribute__ ((progmem))

void int2BCDn(uint16_t bin, lcdpos_t index, char tenpower8);
void int_i2BCDn(int16_t bin, char tenpower8);

// void lcd_data(uint8_t);
void lcd_go2index(lcdpos_t index);
void update_char(char c, lcdpos_t index);
void update_chars(char *c, lcdpos_t index);
uint8_t lcd_cache2comm(uint8_t i);
void busywait(void);

#define lcd_bcd2(x) int2BCDn(x, lcd.next_pos, BCD_DIGIT2)
#define lcd_bcd1(x) update_char('0' + x, lcd.next_pos);

//int printf(const char *format, ...);
void lcd_putstr_P(const char *data);
void lcd_putstr_P_max(const char *data, uint8_t max);

__inline__ void lcd_setpos(lcdpos_t index);
__inline__ lcdpos_t lcd_getpos(void);
uint8_t hex_hi(uint8_t);
uint8_t hex_lo(uint8_t);


extern lcdpos_t lcd_next_pos;   
typedef void(*voidfunc_P)(void);
//const void (PROGMEM *lcd_funcarr[])(void) PROGMEM;

#endif
