/* written by Marcell Gal, cell at x-dsl.hu
 * similar as comm.c but with multichar commands.
 * TODO: control LCD or RS232 VT100 display according to the current submenu
*/
   /*!
   * \addtogroup g_menu Menu Routines
   */
/*@{*/

#include "global.h"
#include "kb.h"
#include "menu.h"
#include "ve.h"
#include "helpers.h"
#include "fuelcontrol.h"
#include "storage.h"
#include "dispatcher.h"
#include "eventqueue.h"
#include "benchmark.h"
#include "comm.h"
#include "lcd.h"
#include "lcd_display.h"
#include "iac.h"
#include "wbo2.h"
#include "spi.h"
#include "knock.h"
#include "boostcontrol.h"
#include "sipr.h"
#include "pid.h"

#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/signal.h>
#include <avr/io.h>
#include <avr/interrupt.h>

#include "ioinit.c"
#include "edis.c"
#include "injector.c"

struct menu_t menu_rs232;

#ifdef PS2KEYBOARD
	struct menu_t menu_ps2;
#endif

#ifdef CAPTURE_ALIEN_ECU
#include "ign_logging.h"
#endif

#include "compile_time.h"
#include "firmware_revision.h"

const char f_version[] PROGMEM = FIRMWARE_VERSION_STR ;

//const char cc[] = __DATE__ " " __TIME__;  //MIK: These symbols could also be used?

// why is this not stored in flash?
static const uint8_t* cfg_bookmarks[]={&config.primep, &config.warmup_clt_range[0], &config.req_fuel,
	&config.tpsdotrate[0], &config.airden_ignore, &config.iac_step_seq,
	&config.ego_conf, &config.wbo2_warmup_ramp, &config.knock_conf,
	&config.ve_learn_coolant, &config.primary_trigger, &config.inj_stage2_rate};

/* decodes a hexa char and writes to the pointed place
returns true if successful 
currently case insensitive 
The _pos_ in the name is historical, it can be used for table values as well, obviously
*/
static __inline__ uint8_t menu_pos_hex(uint8_t c, uint8_t *p)
{
	if(c<='9'){
		if(c>='0'){
			*p = c-'0';
			return 1;
		} else 
			return 0;
	} else {	// c>'9'
		if(c<='F' && c>= 'A'){
			*p = c-('A'-10);
			return 1;
		} else if(c<='f' && c>= 'a'){
			*p = c-('a'-10);
			return 1;
		} else return 0;
	}
}

/* this is very rough, should check m.output for LCD/RS232 
(confuses megatun.. like hell)
and format nicely, etc...
*/
static __inline__ void menu_display(struct menu_t *m, uint8_t val)
{
				tmp_for_comm[0] = 0xd1a1;	// magic number associating for display
				tmp_for_comm[1] = m->posh <<8 | m->posl;	// posh, posl position
				tmp_for_comm[2] = m->table <<8 | val; 	// table and value
				pushfuncpar(&tmp_comm, 2 + 3*8, (void*)tmp_for_comm); // send the 3 words
}

/* 
if write happens to config rpmk or lcd_backlight, the necessary actions 
can be taken
*/
#define conf_offs(a)	((void*)(&config.a) - (void*)(&config))
static inline void write_intercept(uint8_t table, uint8_t offs)
{
	if(table != 'c') 
		return;
		
	switch(offs){
#ifdef LCD
		case(conf_offs(lcd_backlight)):
			lcd_init();
			break;
#endif

#ifdef WBO2
		case(conf_offs(wbo2_nernstdc_target)):
			wbo2_pump_pid_set_target(&pump_pid, config.wbo2_nernstdc_target);
			break;
#endif

		// case :rpmk[0] or rpmk[1] ...
	} // switch
}

/*[0]
checks the position for boundaries.
Writes to the table if looks ok (and returns true)
TODO: notify the display of what's going on
make a new display function that accepts (tablename, line, col, if_success ... )
it should display as much of the actual table as possible 
and update the statusline
*/
static __inline__ uint8_t 
check_write(struct menu_t *m, uint8_t val)
{
	uint8_t posh, posl, offs;
	uint8_t t = m->dimension;

	// TODO: don't do it for table 't' and only horizontally (or take care) for 'b'
	if(m->pos_mode){ // auto positioning
		posh = kpa_idx.uindex;
		posl = rpm_idx.uindex;
	} else {
		posh = m->posh;
		posl = m->posl;
	}
	
	// for automatic mode, it should be always right, but...
	if(posh <= t>>4 && posl <= (t=t&0x0f)){
		offs = posh*(t+1) + posl;
		m->tab[offs] = val;
		write_intercept(m->table, offs);
		// menu_display(m, val);
		// now step to the next col (but not to next line in the j,n,l menus!):
		if(m->posl < t){
			m->posl++;		
		} else 	{ // in the c menu we can usually step to next 'line'
		if(m->table == 'c'){
			if(m->posh < (m->dimension>>4)){
				m->posh++;
				m->posl=0;
			}
		}
		}
		return 1;
	}
	return 0;
}

static inline void mnum(uint8_t action, uint8_t newval){
	uint8_t tmp;
	uint16_t t16;
	switch(action){
		case MNUM_DIGITALOUT:
			cli();
			digitalout(newval, DLOW);
			sei();
			break;
		case MNUM_IGNDWELL:
			engine.igncount = newval;
			cli();
			ign_activate();
			sei_nop_cli();
			readTMR();
			event_insert(etime.now + engine.dwell, IGNDEACT);
			alarm2core();
			sei();
			break;	
		case MNUM_SRLDEBUG:
			pushfuncpar(&srl_generic, DBG_LEN, (void*)((uint16_t)newval << 4));
			break;
		case MNUM_MCB:
			change_baudrate( newval );
			break;
		case MNUM_MSP:
			if(newval > 0x7e){ // higher period NOT supported 
				newval = 0x7e; 
			}
			// save old upper byte:
			tmp = *pointer_to_upper8_16( &debug.stim_period );
			// change upper byte: we only write back 1 byte:
			*pointer_to_upper8_16( &debug.stim_period ) = newval;
			if((tmp == 0) && (newval != 0)){
				// we need the whole, new debug.stim_period here :
				t16 = (newval << 8) || *pointer_to_lower8_16( &debug.stim_period );
				// start an event					
				cli();
				// we could just use event_insert without readTMR() and
				// alarm2core() if we add 0x280 to initial debug.stim_nexttime
				readTMR();
				debug.stim_nexttime = etime.now + t16;
				event_insert(debug.stim_nexttime, STIM_LOW);
				alarm2core();
				sei();
			}
			break;
		case MNUM_MLP:
			lcd_display_setpage( newval );
			break;

	}	// switch(action)
}	// mnum()

/*
Almost the same as check_write()
Changes the current position with the amount specified in the argument.
No auto-increase of position
Dimension and position is taken from the menu_t but tab
pointer must be passed so injection (j) can be modified 
when menu is in ignition (n) or vice versa
*/
static __inline__ uint8_t 
check_modify(struct menu_t *m, uint8_t *tab, int8_t val)
{
	uint8_t posh, posl;
	uint8_t t = m->dimension;

	if(m->pos_mode){ // auto positioning 
		// TODO: modifying just the upper index is not perfect
		posh = kpa_idx.uindex;
		posl = rpm_idx.uindex;
	} else {
		posh = m->posh;
		posl = m->posl;
	}
	
	if(posh <= t>>4 && posl <= (t=t&0x0f)){
		uint16_t n;
		n = tab[posh*(t+1) + posl] + (int16_t)val;

		// we surely don't want a value suddenly overflowing from max to min
		if(n>>8){
			if((n>>8) == 0xFF){ // underflow
				n = 0x00;
			} else { // overflow
				n = 0xff;
			}
		}
		
		tab[posh*(t+1) + posl] = (uint8_t)n;
		// menu_display(m, (uint8_t)n);
		
		return 1;
	}
	return 0;
}

#define TABLE_IN_JN (m->table == 'j' || m->table == 'n' )
#define TABLE_IN_JNL (m->table == 'j' || m->table == 'n' || m->table == 'l')
#define TABLE_IN_JNLD (m->table == 'j' || m->table == 'n' || m->table == 'l' || m->table == 'd')
#define TABLE_IN_CLBT (m->table == 'c' || m->table == 'l' || m->table == 'b' || m->table == 't' )

static __inline__ void enforce_dim(struct menu_t *m)
{
		uint8_t t = m->dimension;
		if(m->posh > t>>4) m->posh = t>>4;
		if(m->posl > (t&0xf)) m->posl = t&0xf;
}

static __inline__ void menu_pos_reset(struct menu_t *m)
{
	m->posh = m->posl = m->pos_mode =  0;
}

// explicite goto mode, we also force a dimension check
static __inline__ void menu_pos_explicit(struct menu_t *m)
{
	m->pos_mode = 0;
//	enforce_dim(m);
}

// the caller must make sure that in correct (j or n) table
static inline void menu_pos_follow(struct menu_t *m)
{
	m->pos_mode = _BV(POS_AUTO);
}

/* default table to use */
static __inline__ void menu_deftab(struct menu_t *m)
{
	m->table = 'j';
	m->tab = ve.ref_table;
	m->dimension = ((VE_SIZE_KPA-1)<<4) + VE_SIZE_RPM - 1;
	enforce_dim(m);
}

void menu_init(struct menu_t *m, uint8_t initstate)
{
	// for keyboard we don't start in megatun.. compatibility state:
    m->state = initstate ;
	menu_pos_reset(m);
	menu_deftab(m);
	// all the rest is fully zero.
};

/* I do not want to clutter the code with case: KEY_ESC
But use ESC, it is catched by the default case anyway..
*/

#define HANDLE_MENU(key, newstate, retvalue) case key: m->state=newstate; return retvalue
#define ACCEPT_MENU(key, newstate) HANDLE_MENU(key, newstate, MENU_RET_OK)
#define REJECT_MENU(key, newstate) HANDLE_MENU(key, newstate, MENU_RET_UNHANDLED)
#define MENU_MNUM(key) m->tab = (uint8_t *) key; m->state = MENU_num;return MENU_RET_OK	
#define MENU_SETVAR(var) m->tab = &var;	m->state = MENU_num; return MENU_RET_OK	

#define	MENU_SET_TABLE_DIM(y, x) (m->dimension = ((y-1)<<4) + x - 1)

void menu_char(struct menu_t * m, uint8_t c)
{
	uint8_t t, pl;
        // special case... if the packet timed out then reset the state
        if ((m->state==MENU_sipr_inc) && (sipr_inc_timeout==255)) {
          // we're not expecting a packet anymore
          m->state=MENU_sipr;
        }
switch(m->state){
    case MENU_ZERO:
		if(c=='M' && comm_state == CMD){	// only act if 'M' is a CMD not a VALUE or OFFSET for comm()
			m->state = MENU_ZERO_M;
			return MENU_RET_OK;
		}
		comm(c);
		return ;
    case MENU_ZERO_M:
		switch(c){
			ACCEPT_MENU('a', MENU_ZERO_Ma);
			default:
				m->state = MENU_ZERO;
				return MENU_RET_UNHANDLED;
		}
    case MENU_ZERO_Ma:
		switch(c){
			case 'n':
				m->state = MENU_mmm;
				pushfuncpar(&srl_generic_P, 8, PSTR("Hello>\r\n"));
				return MENU_RET_OK;
                        case 's':
                                m->state = MENU_sipr;
                                pushfuncpar(&srl_generic_P, 4, PSTR("OK\r\n"));
                                return MENU_RET_OK;
			default:
				m->state = MENU_ZERO;
				return MENU_RET_UNHANDLED;
		}
    case MENU_sipr:     // this is the SIPR protocol wait for packet spot
                switch(c){
                        // see if we get a header packet
                        case SIPR_HEADER:
                                m->state = MENU_sipr_inc;
                                sipr_inc_timeout=2; // allow 100-200ms before timing out on this incoming packet
                                sipr_incoming.idx=0;
                                return MENU_RET_OK;
                        // if we get a M then we'll see if we get a Mas
			case 'M':
				m->state = MENU_sipr_M;
                                return MENU_RET_OK;
                        // anything else gets ignored
                        default:
                                // send a NAK
                                //pushfuncsipr(&sipr_packet_NAK, 6, 0);
                        return MENU_RET_UNHANDLED;
                }
    case MENU_sipr_M:
	switch(c) {
		case 'a':
			m->state = MENU_sipr_Ma;
                        return MENU_RET_OK;
                case SIPR_HEADER:
                        m->state = MENU_sipr_inc;
                        sipr_inc_timeout=2; // allow 100-200ms before timing out on this incoming packet
                        sipr_incoming.idx=0;
                        return MENU_RET_OK;
		default:
			m->state = MENU_sipr;
                        return MENU_RET_OK;
        }
    case MENU_sipr_Ma:
        switch(c) {     
                case 's':
                        m->state = MENU_sipr;
                        pushfuncpar(&srl_generic_P, 4, PSTR("OK\r\n"));
                        return MENU_RET_OK;
                case SIPR_HEADER:
                        m->state = MENU_sipr_inc;
                        sipr_inc_timeout=2; // allow 100-200ms before timing out on this incoming packet
                        sipr_incoming.idx=0;
                        return MENU_RET_OK;
                default:
                        m->state = MENU_sipr;
                        return MENU_RET_OK;
        }
    case MENU_sipr_inc: // processing an incoming SIPR packet
        sipr_process_inc(c);   
        return MENU_RET_OK;
    case MENU_mmm:	// this is the parking menu, most silly commands come back here
		switch(c){
		  // don't use 'M', 'a' or 'n'!
			case KEY_LEFTARROW:
#ifdef LCD
				lcd_display_prevpage();
#endif
				return MENU_RET_OK;

			case KEY_RIGHTARROW:
#ifdef LCD
				lcd_display_nextpage();
#endif
				return MENU_RET_OK;

			case KEY_UPARROW:
#ifdef LCD
				lcd_display_linedec(1);
#endif
				return MENU_RET_OK;

			case KEY_DNARROW:
#ifdef LCD
				lcd_display_lineinc(1);
#endif
				return MENU_RET_OK;

			case KEY_F3:	// people will like the shorthand
#ifdef LCD
				lcd_display_mode(DISPLAY_TABLE, m);
#endif
				m->state = MENU_mtt;
				return MENU_RET_OK;
			case KEY_F11:
#ifdef LCD
			        pushfunc(&lcd_cache2comm, LCD_CHARS+2*LCD_LINES);
#endif
				return MENU_RET_OK;
			ACCEPT_MENU('b', MENU_b);
			ACCEPT_MENU('c', MENU_mc);
			ACCEPT_MENU('d', MENU_md);
			ACCEPT_MENU('i', MENU_mi);
			ACCEPT_MENU('l', MENU_ml);
			ACCEPT_MENU('s', MENU_ms);
			ACCEPT_MENU('t', MENU_mt);
			ACCEPT_MENU('x', MENU_mx);
			ACCEPT_MENU('v', MENU_mv);

			case 'q':
				if (m->table == 'c') check_modify(m, m->tab, 1);
				return MENU_RET_OK;
			case 'w':
				if (m->table == 'c') check_modify(m, m->tab, -1);
				return MENU_RET_OK;
#ifdef DANGEROUS_KNOCK_TESTING
// don't enable this unless you know what you are doing
			case '+':
				if (engine.ignadv_extra < 0xff) engine.ignadv_extra++;
				return MENU_RET_OK;
			case '-':
				if (engine.ignadv_extra > 0) engine.ignadv_extra--;
				return MENU_RET_OK;
			case '*':
				engine.ignadv_extra = 0;
				return MENU_RET_OK;
#endif	// DANGEROUS_KNOCK_TESTING
			default:
				return MENU_RET_UNHANDLED;
		}
    case MENU_mt:
		m->state = MENU_mmm;
		switch(c){
			case 'f':	// fool megatune to modify other table than default ve.ref_table
				if(TABLE_IN_JNL)
					debug.megatune_table = m->tab;
				return MENU_RET_OK;
			case KEY_F3:
			case 't':
#ifdef LCD
				lcd_display_mode(DISPLAY_TABLE, m);	// we hijack the display
#endif
				m->state = MENU_mtt;
				return MENU_RET_OK;
			default:
				return MENU_RET_UNHANDLED;
		}
    case MENU_mtt:
		m->state = MENU_mtt_x;	// this will be next state for most sane input
		switch(c){
			case KEY_UPARROW:
				if(m->posh) m->posh--;
				m->state = MENU_mtt;
				menu_pos_explicit(m);
				return MENU_RET_OK;
			case KEY_DNARROW:
				if(m->posh < m->dimension>>4) m->posh++;
				m->state = MENU_mtt;
				menu_pos_explicit(m);
				return MENU_RET_OK;
			case KEY_LEFTARROW:
				if(m->posl) m->posl--;
				else // in the c menu we can usually step to prev 'line'
					if(m->table == 'c'){
						if(m->posh){
							m->posh--;
							m->posl = 0x0f; // same as m->dimension & 0x0f
						}
					}
				menu_pos_explicit(m);
				m->state = MENU_mtt;
				return MENU_RET_OK;
			case KEY_RIGHTARROW:
				if(m->posl < (m->dimension & 0x0f)) m->posl++;
				else // in the c menu we can usually step to next 'line'
					if(m->table == 'c'){
						if(m->posh < (m->dimension>>4)){
							m->posh++;
							m->posl=0;
						}
					}
				menu_pos_explicit(m);
				m->state = MENU_mtt;
				return MENU_RET_OK;
			//case KEY_F2:    // beware, fall through!
#define BOOKMARK(key, index) \
			case key: \
				m->state=MENU_mtt; \
				if(m->table == 'c') \
				{ uint8_t idx=(uint8_t)(cfg_bookmarks[index]-cfg_bookmarks[0]); \
					m->posh=idx >> 4; \
					m->posl=idx & 0x0f; \
					return MENU_RET_OK; \
				} \
				else return MENU_RET_OK
			BOOKMARK(KEY_F1, 0);
			BOOKMARK(KEY_F2, 1);
			BOOKMARK(KEY_F3, 2);
			BOOKMARK(KEY_F4, 3);
			BOOKMARK(KEY_F5, 4);
			BOOKMARK(KEY_F6, 5);
			BOOKMARK(KEY_F7, 6);
			BOOKMARK(KEY_F8, 7);
			BOOKMARK(KEY_F9, 8);
			BOOKMARK(KEY_F10, 9);
			BOOKMARK(KEY_F11, 10);
			BOOKMARK(KEY_F12, 11);
			case KEY_DEL: // Undo option, loads the current cell into the SRAM
				//TODO:check this why not working?...
				if(m->table == 'c')
					pl=m->posl + (m->posh << 4);
				else
					pl=m->posh*(m->dimension+1) + m->posl;
				get_undovalue(m->table, pl);
				write_intercept(m->table, pl);
				m->state = MENU_mtt;
				return MENU_RET_OK;
			case 'm':	// menu
#ifdef LCD
				lcd_display_mode(DISPLAY_NORMAL, m);	// we give back the display
#endif
				m->state = MENU_mmm;
				return MENU_RET_OK;
			case 'q':	// increase ve (or config element)
				m->state = MENU_mtt;
				if (TABLE_IN_JN) {	// in 'l' ve.lambda will be modified
					if(check_modify(m, ve.ref_table, 1)) // also modify ve.ve_table
						// note that ve.ve_table is increased independently from ve.ref_table.
						ve.ve_table[m->posh*VE_SIZE_RPM + m->posl] += 1<<8;
				} else if (TABLE_IN_CLBT) {
					check_modify(m, m->tab, 1);
				}
				return MENU_RET_OK;
			case 'z':	// decrease ve (or config element)
				m->state = MENU_mtt;
				if (TABLE_IN_JN) { // in 'l' ve.lambda will be modified
					if(check_modify(m, ve.ref_table, -1)) // also modify ve.ve_table
						ve.ve_table[m->posh*VE_SIZE_RPM + m->posl] -= 1<<8;
				} else if (TABLE_IN_CLBT) {
					check_modify(m, m->tab, -1);
				}
				return MENU_RET_OK;
			case 'w':	// increase ign
				m->state = MENU_mtt;
				if (TABLE_IN_JNL)
					check_modify(m, ign.adv_table, 1);
				return MENU_RET_OK;
			case 'x':	// decrease ign
				m->state = MENU_mtt;
				if (TABLE_IN_JNL)
					check_modify(m, ign.adv_table, -1);
				return MENU_RET_OK;
			case 'c':	// config
				m->table = 'c';
				m->tab = (void*)&config;
				m->dimension = (sizeof(struct config_t)-1)| 0x0f;
				menu_pos_explicit(m);
				enforce_dim(m);
				return MENU_RET_OK;
			case 'f':	// goto runtime rpm/kpa follow mode
				if(TABLE_IN_JNL){
					menu_pos_follow(m);
				}
				m->state = MENU_mtt;
				return MENU_RET_OK;
			case 'g':	// goto (position)
				m->state = MENU_mtt_g;
				menu_pos_explicit(m);
				enforce_dim(m);
				return MENU_RET_OK;
#ifdef HWMAP
			case 'h':	// hardware
				m->table = 'h';
				m->tab = hwmap.table;
				MENU_SET_TABLE_DIM(HWMAP_LINES, HWMAP_SIZE);
				enforce_dim(m);
				return MENU_RET_OK;
#endif
			case 'j':	// inj
				menu_deftab(m);
				return MENU_RET_OK;
			case 'l':	// lambda
				m->table = 'l';
				m->tab = ve.lambda;
				MENU_SET_TABLE_DIM(VE_SIZE_KPA, VE_SIZE_RPM);
				enforce_dim(m);
				return MENU_RET_OK;
			case 'k':	//kpa_range
				m->table = 'k';
				m->tab = ve.kpa_range;
				MENU_SET_TABLE_DIM(1, VE_SIZE_KPA);
				menu_pos_explicit(m);
				enforce_dim(m);
				return MENU_RET_OK;
			case 'd':	// ve - ve_ref differences display
				m->state = MENU_mtt;
				menu_deftab(m);
				m->table='d';
				return MENU_RET_OK;
			case 'n':	// ign
				m->table = 'n';
				m->tab = ign.adv_table;
				MENU_SET_TABLE_DIM(VE_SIZE_KPA, VE_SIZE_RPM);
				enforce_dim(m);
				return MENU_RET_OK;
			case 'p':	// goto current runtime rpm/kpa position
				if(TABLE_IN_JNLD){
					m->posh = kpa_idx.uindex;
					m->posl = rpm_idx.uindex;
				}
				m->state = MENU_mtt;
				return MENU_RET_OK;
			case 'r':	//rpm_range
				m->table = 'r';
				m->tab = ve.rpm_range;
				MENU_SET_TABLE_DIM(1, VE_SIZE_RPM);
				menu_pos_explicit(m);
				enforce_dim(m);
				return MENU_RET_OK;
#ifdef BOOSTCONTROL
			case 'b':	// boostmaxtarget_table
				m->table = 'b';
				m->tab = boost.boostmaxtarget_table;
				MENU_SET_TABLE_DIM( BOOSTCONTROL_LINES, VE_SIZE_RPM);
				// menu_pos_explicit(m);	// TODO: needed?
				enforce_dim(m);
				return MENU_RET_OK;
			case 't': // tpstransfer_table
				m->table = 't';
				m->tab = boost.tpstransfer_table;
				MENU_SET_TABLE_DIM( 1, TPSTRANSFERTABLE_SIZE);
				// menu_pos_explicit(m);	// TODO: needed?
				enforce_dim(m);
				return MENU_RET_OK;
#endif
			default:
				m->state = MENU_mtt;
				return MENU_RET_UNHANDLED;
		}
    case MENU_mtt_g:
		if(menu_pos_hex(c, &m->num))
			m->state = MENU_mtt_gx;	//great, received valid hexa num 
		else
			m->state = MENU_mtt;	// anything silly is escape
		return;
	case MENU_mtt_gx:
		m->state = MENU_mtt;
		if(menu_pos_hex(c, &m->posl)){	// good, valid char
			m->posh = m->num;
			menu_pos_explicit(m);	// explicite mode: only for valid hexa char
			enforce_dim(m);
			return MENU_RET_OK;
		}
		return MENU_RET_UNHANDLED;
	case MENU_b:
		switch(c){
			case 'y':
				m->state = MENU_by;
				return MENU_RET_OK;
			default:
				m->state = MENU_mmm;
				return MENU_RET_UNHANDLED;
		}
    case MENU_by:
		switch(c){
			case 'e':
				m->state = MENU_ZERO;	// leaving 'Man'ual mode for rs232 is ok. for ps2keyboad quite meaningless
				pushfuncpar(&srl_generic_P, 5, PSTR("Bye\r\n"));
				return MENU_RET_OK;
			default:
				m->state = MENU_mmm;
				return MENU_RET_UNHANDLED;
		}
	case MENU_mtt_x:	// note that this is used for any tablevalue with 2 hexachar length
		// TODO: handle '+' or '-' as in the wiki
		if(menu_pos_hex(c, &m->num))
			m->state = MENU_mtt_xx;	//great, received valid hexa num 
		else
			m->state = MENU_mtt;	// anything silly is escape
		return;
	case MENU_mtt_xx: // higher hexdigit is in num, tablename in m->table
		m->state = MENU_mtt;
		if(menu_pos_hex(c, &t)){ // great, received valid hexa num 
			t |= m->num << 4;	// retreive the higher hexdigit of value
			// now we check which table was on table (hehehe :-) and act accordingly
			switch(m->table){
			case 'c':
				check_write(m, t);
				return;
			case 'j':	// write to both ve.table_ref and the autotuned table:
				pl = m->posl; // check_write increments it..
				if(check_write(m, t)){
					ve.ve_table[m->posh*VE_SIZE_RPM + pl] = (t<<8)|127;
				}
				return;
			case 'k':
				check_write(m, t);
				return;
			// case 'd': do nothing (just for display)...
			case 'l':
#ifdef BOOSTCONTROL
			case 'b':
			case 't':
#endif
				check_write(m, t);
				return;
#ifdef HWMAP
			case 'h':
				check_write(m, t);
				return;
#endif
			case 'n':
				check_write(m, t);
				return;
			case 'r':
				check_write(m, t);
				return;
			default:
				return;
			} // switch
		return;
		} // if valid char
	case MENU_mx:	// special menu for experiments
		switch(c){
			// mxa and mxA gone, check mde.. bit7
			case 'c':	// commtest
				pushfuncpar(&srl_generic_P, 25, PSTR("Yes, menu and comm worx\r\n"));
				return MENU_RET_OK;
			case 'i':	// init iac
				init_iac();
				return MENU_RET_OK;	
			case 'l':	// deprecated. IT WILL GO! USE mll
#ifdef LCD
				pushfunc(&lcd_cache2comm, LCD_CHARS+2*LCD_LINES);
#endif
				return MENU_RET_OK;
#ifdef GENBOARDv3
		case 'q':
			set_knock_channel(KNOCK_CHANNEL_1);
			return MENU_RET_OK;
		case 'Q':
			set_knock_channel(KNOCK_CHANNEL_2);
			return MENU_RET_OK;
		case 'z':
			read_knock();
			return MENU_RET_OK;
		case 'Z':
			pushfuncpar(&srl_generic, sizeof(knock.data), (uint8_t *)&knock.data);
			return MENU_RET_OK;
#endif			

#ifdef PS2KEYBOARD
		case 'k':  // I'm lost in menu_ps2, tell me where I am!
			pushfuncpar(&srl_generic, sizeof(struct menu_t), (uint8_t *)&menu_ps2);
			return MENU_RET_OK;
#endif
		case 'd':   // dump ps2 menu char
			debug_ps2 ^= 0x02;
			return MENU_RET_OK;
		case 'o':	// escape the second mxo.. to get the effect
#ifndef GENBOARDv3  // don't fiddle with injector pwm'ing
				if(debug.ocr0_val >= 0x20){	// 249 is 1kHz, below 32 would be a flood...
#ifndef EVENTQUEUE_TCNT2
					TCCR0 =  _BV(WGM01) | _BV(COM01) | _BV(CS02) | _BV(CS01);
#endif
					OCR0 = debug.ocr0_val;
					TIMSK |= _BV(OCIE0);
				} else { // set back 4096 usec period...
					TCCR0 =  _BV(WGM00) | _BV(WGM01) | _BV(COM01) | _BV(CS02) | _BV(CS01);
					TIMSK &= ~_BV(OCIE0);
				}
				m->tab = &debug.ocr0_val;
				m->state = MENU_num;
#endif
				return MENU_RET_OK;
		case 's':  // dump ps2 scancode
			debug_ps2 ^= 0x01;
			return MENU_RET_OK;
#ifdef CAPTURE_ALIEN_ECU
		case 'L': 
			if((*comm_TX_plainbytes == 0) && ! comm_buffer_full()){
				alien_mapsensor_log();
			}
			return MENU_RET_OK;
#endif
		case 't':
#ifdef BENCHMARK
			benchstats.lowest_rpm_p=0xFFFFFFFF;
			benchstats.highest_rpm_p=0;
			return MENU_RET_OK;
#endif
#ifdef WBO2
		case 'h': MENU_SETVAR( wbo2_heater_pwm.designated_ratio );// wbo2 heater dutycycle
#endif
			default:
				m->state = MENU_mmm;
				return MENU_RET_UNHANDLED;
		}
	case MENU_ml:	// special menu for lcd experiments
		switch(c){
#ifdef LCD		
			case 'a':
				lcd_start_address_test();
				return MENU_RET_OK;
			case 'd':
				lcd_dirty();
				return MENU_RET_OK;
			case 'l':
				pushfunc(&lcd_cache2comm, LCD_CHARS+2*LCD_LINES);
				return MENU_RET_OK;
			case 'i':
				lcd_init();
				return MENU_RET_OK;
			case 'p':	// selects LCD page by number (see PS2 leftarrow / rightarrow)
				MENU_MNUM( MNUM_MLP );
				return MENU_RET_OK;
			case 's': // you need to do mlsxx mlsxx (twice!) switchoff: mlsff mlsff 
				if(lcd_disp.param != 0xFF){
					lcd_display_mode(DISPLAY_HOLD, m);
					lcd_charseq_test(lcd_disp.param);
				} else {
					lcd_display_mode(DISPLAY_NORMAL, m);
				}
				m->tab = &lcd_disp.param;
				m->state = MENU_num;
				return MENU_RET_OK;
#endif
			default:
				m->state = MENU_mmm;
				return MENU_RET_UNHANDLED;
		}
	case MENU_md:	// special menu for debugging
		switch(c){
			case 'a': MENU_SETVAR( debug.forced_adv );
			case 'd': MENU_SETVAR( debug.disable_flags );	// disable features
			case 'e': MENU_SETVAR(	debug.enable_flags );	// enable features
			case 'f': MENU_SETVAR( debug.auto_dump_freq );	// comm_auto_dump frequency
			case 'h': MENU_MNUM( MNUM_DIGITALOUT ); // hardware test: set low state under 0x80, high state above
			case 'i': MENU_SETVAR( debug.forced_iac );	
			case 'n': MENU_MNUM( MNUM_IGNDWELL );	// test coil sequence
			case 'j': // TODO: test injection sequence (mdh not enough?)
				return MENU_RET_OK;	
			case 'k': MENU_SETVAR( debug.auto_dump_mask );
#ifdef GENBOARDv3
			case 'K':
				knock.data = 0;  // start sampling if debug mode enabled
				m->state = MENU_mmm;
				return MENU_RET_OK;
#endif
			case 'v':
				pushfuncpar(&srl_generic_P, sizeof(c_time), c_time);
				m->state = MENU_mmm;
				return MENU_RET_OK;
                        // query firmware version number
			case 'V':
				pushfuncpar(&srl_generic_P, sizeof(f_version), f_version);
				m->state = MENU_mmm;
				return MENU_RET_OK;
			case 'r':
				pushfuncpar(&srl_generic, sizeof(struct engine_t), (uint8_t *)&engine);
				m->state = MENU_mmm;
				return MENU_RET_OK;
			case 's':  // raw sensor readings
				pushfuncpar(&srl_generic, SENSORS_BYTES, (void*)&sensor_filt);
				m->state = MENU_mmm;
				return MENU_RET_OK;
			case 'p': MENU_SETVAR( debug.forced_pw );
			case 'x': MENU_MNUM( MNUM_SRLDEBUG );				
			default:
				m->state = MENU_mmm;
				return MENU_RET_UNHANDLED;
		}
	case MENU_num:
			if(menu_pos_hex(c, &m->num)){
				m->state = MENU_num_x;
				return MENU_RET_OK;
			} else {
				menu_deftab(m);
				m->state = MENU_mmm;
				return MENU_RET_UNHANDLED;
			}
	case MENU_num_x:
			m->state = MENU_mmm;	// we do not store prev_state, so fall back to mmm
			if(menu_pos_hex(c, &t)){ // good, valid char
				uint8_t *p = m->tab;
				t |= m->num << 4;
				// low m->tab values are handled specially:
				if((((uint16_t) p)>> 8) == 0){
					mnum((uint16_t) p  & 0xff, t);
					return MENU_RET_OK;				
				}
				
				// writing 1 byte never needs cli():
				*m->tab = t;
				menu_deftab(m);
				return MENU_RET_OK;
			}
			menu_deftab(m);
			return MENU_RET_UNHANDLED;
	case MENU_mc:	// menu for config
		m->state = MENU_mmm;
		switch(c){
			case 'b': // bootloader; TODO: kill debug.bootl and use MNUM
				if(debug.bootl != 0){ // mcb00 == last point of return!
				  if (!(engine.status & _BV(running)) || (debug.bootl == 0xfe)) {
				    cli();
				    if (debug.bootl >= 0xfe) {	// fe is forced, even running input trigger
				      SPCR = 0x35; // bootloader takes care of io
				    } else {
				      SPCR = 0x36; // we take care of io and uart
				      ioinit();
					  change_baudrate( debug.bootl );
				      UCSR0A = 0;
				      UCSR0B = 0x18; //d8;
				      UCSR0C = 0x06;
				    }
				    // jump to the last address in progmem
				    asm("jmp 0x1fffc"); // bye bye cruel world I'm leaving
				  }
				  debug.bootl=0; // in case engine was running
				}
				m->tab = &debug.bootl;
				m->state = MENU_num;
				return MENU_RET_OK;
			case 'B': MENU_MNUM( MNUM_MCB );	// change baudrate
			case 'd': // dump config
				pushfuncpar(&ConfigDump, 0xFF, NULL);
				return MENU_RET_OK;
			case 's': // save all variables			
				save_variables();
#ifdef LCD
				lcd_init(); // Just to see: we entered the correct key sequence
#endif
				return MENU_RET_OK;
/*
  TODO: add command that returns "eeprom_flags & _BV(active)", then
	the box can be polled for when eeprom write has completed
*/
			case 't':
				pushfuncpar(&TableDump, 0xFF, (void*)(('j'<<8)+ VE_SIZE_KPA-1));
				pushfuncpar(&TableDump, 0xFF, (void*)(('l'<<8)+ VE_SIZE_KPA-1));
				pushfuncpar(&TableDump, 0xFF, (void*)(('n'<<8)+ VE_SIZE_KPA-1));
				pushfuncpar(&TableDump, 0xFF, (void*)('k'<<8));
				pushfuncpar(&TableDump, 0xFF, (void*)('r'<<8));
#ifdef HWMAP
				pushfuncpar(&TableDump, 0xFF, (void*)('h'<<8)+(HWMAP_LINES-1));
#endif

#ifdef BOOSTCONTROL
				pushfuncpar(&TableDump, 0xFF, (void*)('b'<<8)+(BOOSTCONTROL_LINES-1));
				pushfuncpar(&TableDump, 0xFF, (void*)('t'<<8)+(1-1));
#endif
				return MENU_RET_OK;
			case 'i': // talk about yourself
				// the identification string lives 12
				// bytes below the bootloader jmp-instruction. 
				// Flash is indexed by words
				pushfuncpar(&srl_generic_P_upper, 12, (uint8_t*)(0xfffc-12) );
				m->state = MENU_mmm;
				return MENU_RET_OK;
			default:
				
				return MENU_RET_UNHANDLED;
		}
	case MENU_mv:	// ve autotune menu
		m->state = MENU_mmm;
		switch(c) {
			case 'r': // resets learning
				copyVEreftoVE();  // BEWARE, THIS ERASES ALL WE HAVE LEARNED!			
				return MENU_RET_OK;
			case 's': // save to ve_ref what we have learnt (not to eeprom, see mca)
				copyVEtoVEref();
				return MENU_RET_OK;
			default:
				m->state = MENU_mmm;
				return MENU_RET_UNHANDLED;
		}
	case MENU_ms:	// stimulation menu
		switch(c){
		// when nonzero, put a starting event into menu, it can easily become 2 or more interfering 
		// (and cpu exhausting) sets of output events. To avoid this,
		// switch off stimulation with p=00 not switch from one period 
		// to the other: used like msp34msp34msp00msp00msp67msp67
			case 'p':	// period upper byte (1024 usec)
				MENU_MNUM( MNUM_MSP );
			case 'q':	// period lower byte (4 usec)
				m->tab = pointer_to_lower8_16(&debug.stim_period);
				m->state = MENU_num;
				return MENU_RET_OK;				
			case 't':	// packed: 7:6 is the missing teeth to 0..3; 5:0 is toothcount 0..63
				m->tab = &debug.stim_tooth;
				m->state = MENU_num;
				return MENU_RET_OK;
			default:
				m->state = MENU_mmm;
				return MENU_RET_UNHANDLED;
			}
	default:	// should never reach
		m->state = MENU_mmm;
}; // switch(m->state)


}	// menu_char

/*@}*/

