   /*!
   * \addtogroup g_helpers Specialized Math Routines
   */
/*@{*/

#ifndef __PID_C__
#define __PID_C__
/* Copyright Michael Kristensen, 2003
	All rights reserved.
	Distribution or modification is only possible according to 
	the GenboardPublicLicense.txt, which you should have recieved
	with this file. If it's for some reason not available, or you
	do not fully apply the conditions,
	Distribution or modification is disallowed.
*/

#include "eventqueue.h"
#include "pid.h"
#include "helpers.h"
#include "benchmark.h"
#include <inttypes.h>

#include "arithm.c"

// cast variables from config_t to a pid_parameters_t type
void pid_init(struct pid_t *p, struct pid_parameters_t *parameters) {
	p->parameters = parameters;

}

void pid_set_target(struct pid_t *p, uint16_t target) {
	p->target = target;
}

// NOTE: target/current should be 15 bit max to prevent overflow
int16_t pid15(struct pid_t *p, uint16_t current) {
	int16_t error, integral, dot, res, i16;
	int32_t i;
	struct pid_parameters_t *pp;
	uint8_t a, b;

	// error term and overflow check
	a = p->target >> 8;
	b = current >> 8;
	res = (uint16_t)a - (uint16_t)b;
	if (res > 127) {
		error = 32767;
#ifdef BENCHMARK
		benchstats.pid_overflow++;
#endif
	} else if (res < -128) {
		error = -32768;
#ifdef BENCHMARK
		benchstats.pid_overflow++;
#endif
	} else {
		error = p->target - current;
	}
	
	// A positive change results in negative d(). This is caused by the way 'error' is defined
	dot = p->prev - current;
	p->prev = current;

	// integrate error term
	// TODO: we need "anti-windup"
	integral = p->integral;
	i = (int32_t)integral + error;

	// handle overflow
	i16 = i>>8;
	if (i16 > 127)
		res = 32767;
	else if (i16 < -128)
		res = -32768;
	else
		res = i;
	
	pp = p->parameters;

	// yes please, we really want to limit the integrator
	i16 =  (int16_t)pp->integral_limit << 6;  // dont shift more than 7!
	integral = limit_i16(res, i16, -i16);

	p->integral = integral;

	res = mult16i_8_div256(error, pp->kp) +
		mult16i_8_div256(integral, pp->ki) +
		mult16i_8_div256(dot, pp->kd);

	return res;
}

// NOTE: target/current should be 15 bit max to prevent overflow
// Soft deadband around target where integration is slowed down
int16_t pid15_soft(struct pid_t *p, uint16_t current) {
	int16_t error, integral, dot, res, i16;
	uint16_t ptarget, errorabs;
	int32_t i;
	struct pid_parameters_t *pp;
	uint8_t a, b;

	// error term and overflow check
	ptarget = p->target;
	a = ptarget >> 8;
	b = current >> 8;
	res = (uint16_t)a - (uint16_t)b;
	if (res > 127) {
		error = 32767;
#ifdef BENCHMARK
		benchstats.pid_overflow++;
#endif
	} else if (res < -128) {
		error = -32767;
#ifdef BENCHMARK
		benchstats.pid_overflow++;
#endif
	} else {
		error = ptarget - current;
	}
	
	// A positive change results in negative d(). This is caused by the way 'error' is defined
	dot = p->prev - current;
	p->prev = current;

	// integrate error term
	// TODO: we need "anti-windup"
	integral = p->integral;
	
	if(error<0)
		errorabs = 0 - error;
	else
		errorabs = error;

	/* this is a large window: 0.5 .. 1.5*ptarget, 
		probably good for Ri. 
		We could append a config byte to the end of
		pid_parameters_t which wouldn't be used for the
		primitive PID controller 
	*/
	if(errorabs < (ptarget >>1))
		i = (int32_t)integral + (error>>2);
	else
		i = (int32_t)integral + error;

	// handle overflow
	i16 = i>>8;
	if (i16 > 127)
		res = 32767;
	else if (i16 < -128)
		res = -32768;
	else
		res = i;
	
	pp = p->parameters;

	// yes please, we really want to limit the integrator
	i16 =  (int16_t)pp->integral_limit << 6;  // dont shift more than 7!
	integral = limit_i16(res, i16, -i16);

	p->integral = integral;

	res = mult16i_8_div256(error, pp->kp) +
		mult16i_8_div256(integral, pp->ki) +
		mult16i_8_div256(dot, pp->kd);

	return res;
}


#endif
